/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2003, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __POPULATION__
#include "population.h"
#endif

#ifndef __MATHUTILS__
#include "mathutils.h"
#endif

const unsigned kDefaultSize = 32;
const unsigned kDefaultSurvivors = 0;

Population::Population( Genome *inSeed )
: mSize(0), mSurvivors(kDefaultSurvivors), mGenome(0), mSelectByRank(false)
{
	mSeed = inSeed;
}

Population::~Population()
{
	if( mSeed )
		delete mSeed;

	if( mGenome )
	{
		for(unsigned i=0;i<mSize;i++)
			if( mGenome[i] )
				delete mGenome[i];
		delete[] mGenome;
	}

	if( mChildren )
		delete mChildren;
}


// set the size of the population
void Population::SetSize( unsigned inSize )
{
	Genome **ng = new Genome *[inSize];
	mChildren = new Genome *[inSize]; // max # of children regardless of mSurvivors
	if( ng )
	{
		unsigned i;
		if( mGenome )
		{
			// delete any extras
			for(i=inSize;i<mSize;i++)
			{
				if( mGenome[i] )
					delete mGenome[i];
			}

			// copy all left
			for(i=0;i<inSize;i++)
			{
				if( i < mSize )
					ng[i] = mGenome[i];
				else
					ng[i] = 0; 
			}

			delete[] mGenome;
		}
		else
		{
			for(i=0;i<inSize;i++)
				ng[i] = 0; 
		}

		mSize = inSize;
		mGenome = ng;
	}
}

// set the number of survives kept unchanged each cycle, must be no greater than mSize
void Population::SetSurvivors( unsigned inSurvivors )
{
	mSurvivors = inSurvivors;
}

// seed initial population
void Population::Seed(bool inForceToZero)
{
	if( !mSize )
		SetSize( kDefaultSize ); // default the size
	if( mGenome )
	{
		for(unsigned i=0;i<mSize;i++)
		{
			if( mGenome[i] )
				delete mGenome[i];
			mGenome[i] = 0;
			mGenome[i] = mSeed->Spawn();
			if( inForceToZero )
				mGenome[i]->Zero();
			else
				mGenome[i]->Random();
			mGenome[i]->CalcFitness();
		}

		Sort();
	}
}

// evolve the population one generation forward
void Population::Evolve(void)
{
	// first off, sort the current population,
	// because several steps later on require this.
	Sort();
	
	// find out the crossover probability to use.
	double crossoverProb = 0.0;
	if( mSize )
		crossoverProb = mGenome[0]->GetBlueprint()->GetCrossoverProb();

	// crossover 
	if( mSize > mSurvivors )
	{
		unsigned i, children;
		children = mSize - mSurvivors;
		for(i=0;i<children;i++)
			mChildren[i] = 0; // null out the pointers

		for(i=0;i<children;i++)
		{
			Genome *mommy, *daddy;

			// do either copy or crossover,
			// according to crossover probability.
			if( Random1() < crossoverProb )
			{
				SelectParents(&mommy,&daddy);
				mChildren[i] = mommy->Crossover(daddy);
			}
			else
			{
				SelectParents(&mommy,NULL);
				mChildren[i] = mommy->Copy();
			}
			mChildren[i]->Mutate();
			// we have to do this even if we copied in case there was a mutation,
			// we could make Mutate() return a boolean, but that would require more work
			// out of the inherited function which is usually something best avoided.
			mChildren[i]->CalcFitness();
		}

		// replace the losers from the last generation
		// with the new children
		for(i=0;i<children;i++)
		{
			if( mGenome[mSurvivors+i] )
				delete mGenome[mSurvivors+i];
			mGenome[mSurvivors+i] = mChildren[i];
		}
	}
}

// this is used by MultiPopulation, assumes inGenome->CalcFitness has already been called.
Genome *Population::SwapGenome(unsigned index, Genome *inGenome)
{
	Genome *g = mGenome[index];
	mGenome[index] = inGenome;
	return g;
}

// get index of winner
unsigned Population::GetWinner(void) const
{
	// find the most fit genome and return its index
	double best;
	unsigned bestIndex = 0;

	if( mGenome && mSize > 0 )
	{
		best = mGenome[0]->GetFitness();

		for(unsigned i=1;i<mSize;i++)
		{
			double fitness = mGenome[i]->GetFitness();
			if( fitness > best )
			{
				best = fitness;
				bestIndex = i;
			}
		}
	}

	return bestIndex;
}

static int CompareGenomes(const void *a, const void *b)
{
	const Genome *ag = *(const Genome **)a, *bg = *(const Genome **)b;
	double diff = bg->GetFitness() - ag->GetFitness();
	if( diff == 0.0 )
		return 0;
	else if( diff < 0.0 )
		return -1;
	else // > 0.0
		return 1;
}

// sort population by fitness which must have already been calculated.
void Population::Sort(void)
{
	// Just using qsort
	// ::qsort(base,number,size,proc)
	::qsort( (void *)mGenome, mSize, sizeof(Genome *), CompareGenomes );

}

unsigned Population::SelectParent( void )
{
	if( mSelectByRank )
		return SelectParentByRank();
	else
		return SelectParentByFitness();
}

// override this to change selection method, default is by rank,
// return index of selected parent
unsigned Population::SelectParentByRank( void )
{
	// average WFE on walkforward test of EMA cross is 44.9093

	int total = 0, rank, r;
	unsigned index;

	for(index=0;index<mSize;index++)
		total += (index+1);

	r = rand() % total;

	rank = mSize;
	index = 0;
	// had a huge bug here that made return always zero.
	while( index < mSize )
	{
		r -= rank;
		
		if( r < 0 )
			break; // current index is the one.

		rank--;
		index++;
	}
	if( index >= mSize )
		index = mSize - 1;

	return index;
}


// select parent by fitness
unsigned Population::SelectParentByFitness( void )
{
	// average WFE in walkforward test of 8.5785%

	double total = 0.0, rank, r, minRank;
	unsigned index;

	// find preliminary total
	for(index=0;index<mSize;index++)
		total += mGenome[index]->GetFitness();
	
	if( total == 0.0 )
	{
		total = (double)mSize; // special case for when all fitnesses are zero.
	}

	// don' let anything have a fitness of zero for selection purposes.
	minRank = total / (double)(mSize*100);

	// recalculate total with that taken into account.
	for(index=0;index<mSize;index++)
	{
		rank = mGenome[index]->GetFitness();
		if( rank <= minRank )
			rank = minRank;

		total += mGenome[index]->GetFitness();
	}


	r = Random1() * total;

	index = 0;
	while( index < mSize )
	{
		rank = mGenome[index]->GetFitness();
		if( rank <= minRank )
			rank = minRank;

		r -= rank;
		
		if( r < 0.0 )
			break; // current index is the one.

		index++;
	}
	if( index >= mSize )
		index = mSize - 1;

	return index;
}

	
// genomes will already have had their fitness values calculated.
void Population::SelectParents( Genome **outMother, Genome **outFather )
{
	unsigned mother, father;

	father = mother = SelectParent();
	*outMother = mGenome[mother];
	
	// they may be doing a copy instead of a crossover and only need one parent.
	if( outFather )
	{
		// we had a huge bug here that made crossover parents always the same
		if( mSize > 1 )
		{
			while(father == mother )
				father = SelectParent();
		}

		*outFather = mGenome[father];
	}
}


